/*
*/

const Cc = Components.classes;
const Ci = Components.interfaces;

const PREF_EM_UPDATE_ENABLED          = "extensions.update.enabled";

Components.utils.import("resource://gre/modules/Services.jsm");

var EXPORTED_SYMBOLS = [ "AddonManager", "AddonManagerInternal" ];

const PROVIDERS = [
  "resource://gre/modules/XPIProvider.jsm",
  "resource://gre/modules/LightweightThemeManager.jsm"
];

function LOG(str) {
  dump("*** addons.manager: " + str + "\n");
}

// Calls a callback function and consumes any exception.
function safeCall(callback) {
  var args = Array.slice(arguments, 1);
  try {
    return callback.apply(null, args);
  }
  catch (e) {
    LOG("Exception calling callback: " + e);
  }
}

/**
 * This is the real manager, kept here rather than in AddonManager to keep its
 * contents hidden from API users.
 */
var AddonManagerPrivate = {
  installListeners: null,
  addonListeners: null,
  providers: [],
  started: false,

  startup: function AM_startup() {
    if (this.started)
      return;

    this.installListeners = [];
    this.addonListeners = [];

    let appChanged = true;
    try {
      appChanged = Services.appinfo.version != Services.prefs.getCharPref("extensions.lastAppVersion");
    }
    catch (e) { }
    if (appChanged) {
      LOG("Application has been upgraded");
      Services.prefs.setCharPref("extensions.lastAppVersion", Services.appinfo.version);
    }

    // Ensure all known providers have had a chance to register themselves
    PROVIDERS.forEach(function(url) {
      Components.utils.import(url, {});
    });

    let needsRestart = false;
    this.providers.forEach(function(provider) {
      if ("startup" in provider)
        provider.startup();
      if (("checkForChanges" in provider) && provider.checkForChanges(appChanged))
        needsRestart = true;
    });
    this.started = true;

    if (needsRestart) {
      let appStartup = Cc["@mozilla.org/toolkit/app-startup;1"].
                       getService(Ci.nsIAppStartup2);
      appStartup.needsRestart = needsRestart;
    }
  },

  registerProvider: function AM_registerProvider(provider) {
    this.providers.push(provider);

    // If we're registering after startup call this providers startup immediately.
    if (this.started && "startup" in provider)
      provider.startup();
  },

  shutdown: function AM_shutdown() {
    this.providers.forEach(function(provider) {
      if ("shutdown" in provider)
        provider.shutdown();
    });

    this.installListeners = null;
    this.addonListeners = null;
    this.started = false;
  },

  backgroundUpdateCheck: function AM_backgroundUpdateCheck() {
    if (!Services.prefs.getBoolPref(PREF_EM_UPDATE_ENABLED))
      return;
    this.getAddonsByType(null, function(addons) {
      addons.forEach(function(addon) {
        if (addon.permissions & AddonManager.PERM_CANUPGRADE) {
          addon.findUpdates({
            onUpdateAvailable: function(addon, install) {
              install.install();
            }
          }, AddonManager.UPDATE_WHEN_PERIODIC_UPDATE);
        }
      });
    });
  },

  callInstallListeners: function AM_callInstallListeners(method, additional) {
    let result = true;
    let listeners = this.installListeners;
    if (additional)
      listeners = additional.concat(listeners);
    let args = Array.slice(arguments, 2);
    listeners.forEach(function(listener) {
      try {
        if (method in listener) {
          if (listener[method].apply(listener, args) === false)
            result = false;
        }
      }
      catch (e) { }
    });
    return result;
  },

  callAddonListeners: function AM_callInstallListeners(method) {
    var args = Array.slice(arguments, 1);
    this.addonListeners.forEach(function(listener) {
      try {
        if (method in listener)
          listener[method].apply(listener, args);
      }
      catch (e) { }
    });
  },

  notifyAddonSelected: function AM_notifyAddonSelected(id, type, pendingRestart) {
    this.providers.forEach(function(provider) {
      provider.addonSelected(id, type, pendingRestart);
    });
  },

  getInstallForURL: function AM_getInstallForURL(url, callback, mimetype, hash, name, iconURL, version, loadgroup) {
    if (!url || !mimetype || !callback)
      throw new TypeError("Invalid arguments");

    for (let i = 0; i < this.providers.length; i++) {
      if ("supportsMimetype" in this.providers[i] &&
          this.providers[i].supportsMimetype(mimetype)) {
        this.providers[i].getInstallForURL(url, hash, name, iconURL, version, loadgroup, function(install) {
          safeCall(callback, install);
        });
        return;
      }
    }
    callback(null);
  },

  getInstallForFile: function AM_getInstallForFile(file, callback, mimetype) {
    if (!file || !callback)
      throw new TypeError("Invalid arguments");

    // Freeze the list of providers then call each in turn till one gives a
    // valid Install
    let providers = this.providers.slice(0);

    function callNextProvider() {
      if (providers.length == 0)
        return safeCall(callback, null);

      let provider = providers.shift();
      if ("getInstallForFile" in provider) {
        provider.getInstallForFile(file, function(install) {
          if (install)
            safeCall(callback, install);
          else
            callNextProvider();
        });
      }
      else {
        callNextProvider();
      }
    }

    callNextProvider();
  },

  getInstalls: function AM_getInstalls(types, callback) {
    if (!callback)
      throw new TypeError("Invalid arguments");

    // Freeze the list of providers then call each in turn to get their Installs
    let providers = this.providers.slice(0);
    let installs = [];

    function callNextProvider() {
      if (providers.length == 0)
        return safeCall(callback, installs);

      let provider = providers.shift();
      if ("getInstalls" in provider) {
        provider.getInstalls(types, function(providerInstalls) {
          installs = installs.concat(providerInstalls);
          callNextProvider();
        });
      }
      else {
        callNextProvider()
      }
    }

    callNextProvider();
  },

  isInstallEnabled: function AM_isInstallEnabled(mimetype) {
    for (let i = 0; i < this.providers.length; i++) {
      if ("supportsMimetype" in this.providers[i] &&
          this.providers[i].supportsMimetype(mimetype))
        return this.providers[i].isInstallEnabled();
    }
  },

  isInstallAllowed: function AM_isInstallAllowed(mimetype, uri) {
    for (let i = 0; i < this.providers.length; i++) {
      if ("supportsMimetype" in this.providers[i] &&
          this.providers[i].supportsMimetype(mimetype))
        return this.providers[i].isInstallAllowed(uri);
    }
  },

  startInstallation: function AM_startInstallation(mimetype, source, uri, installs) {
    if (!("@mozilla.org/extensions/web-install-listener;1" in Cc)) {
      LOG("No web installer available, cancelling all installs");
      installs.forEach(function(install) {
        install.cancel();
      });
      return;
    }

    try {
      let weblistener = Cc["@mozilla.org/extensions/web-install-listener;1"].
                        getService(Ci.extIWebInstallListener);

      if (!this.isInstallAllowed(mimetype, uri)) {
        if (weblistener.onWebInstallBlocked(source, uri, installs, installs.length)) {
          installs.forEach(function(install) {
            install.install();
          });
        }
      }
      else if (weblistener.onWebInstallRequested(source, uri, installs, installs.length)) {
        installs.forEach(function(install) {
          install.install();
        });
      }
    }
    catch (e) {
      LOG("Failure calling web installer: " + e);
      installs.forEach(function(install) {
        install.cancel();
      });
      return;
    }
  },

  addInstallListener: function AM_addInstallListener(listener) {
    this.removeInstallListener(listener);
    this.installListeners.push(listener);
  },

  removeInstallListener: function AM_removeInstallListener(listener) {
    this.installListeners = this.installListeners.filter(function(i) {
      return i != listener;
    });
  },

  getAddon: function AM_getAddon(id, callback) {
    if (!id || !callback)
      throw new TypeError("Invalid arguments");

    // Freeze the list of providers then call each in turn till one gives a
    // valid Install
    let providers = this.providers.slice(0);

    function callNextProvider() {
      if (providers.length == 0)
        return safeCall(callback, null);

      providers.shift().getAddon(id, function(addon) {
        if (addon)
          safeCall(callback, addon);
        else
          callNextProvider();
      });
    }

    callNextProvider();
  },

  getAddons: function AM_getAddon(ids, callback) {
    if (!ids || !callback)
      throw new TypeError("Invalid arguments");

    // Clone the array to avoid breaking the caller
    ids = ids.slice(0);
    let addons = [];

    function getNextAddon() {
      if (ids.length == 0)
        return safeCall(callback, addons);

      AddonManagerPrivate.getAddon(ids.shift(), function(addon) {
        addons.push(addon);
        getNextAddon();
      });
    }

    getNextAddon();
  },

  getAddonsByType: function AM_getAddonsByType(types, callback) {
    if (!callback)
      throw new TypeError("Invalid arguments");

    // Freeze the list of providers then call each in turn to get their Installs
    let providers = this.providers.slice(0);
    let addons = [];

    function callNextProvider() {
      if (providers.length == 0)
        return safeCall(callback, addons);

      providers.shift().getAddonsByType(types, function(providerAddons) {
        addons = addons.concat(providerAddons);
        callNextProvider();
      });
    }

    callNextProvider();
  },

  getAddonsWithPendingOperations: function AM_getAddonsWithPendingOperations(types, callback) {
    if (!callback)
      throw new TypeError("Invalid arguments");

    // Freeze the list of providers then call each in turn to get their Installs
    let providers = this.providers.slice(0);
    let addons = [];

    function callNextProvider() {
      if (providers.length == 0)
        return safeCall(callback, addons);

      let provider = providers.shift();
      if ("getAddonsWithPendingOperations" in provider) {
        provider.getAddonsWithPendingOperations(types, function(providerAddons) {
          addons = addons.concat(providerAddons);
          callNextProvider();
        });
      }
      else {
        callNextProvider();
      }
    }

    callNextProvider();
  },

  addAddonListener: function AM_addAddonListener(listener) {
    this.removeAddonListener(listener);
    this.addonListeners.push(listener);
  },

  removeAddonListener: function AM_removeAddonListener(listener) {
    this.addonListeners = this.addonListeners.filter(function(i) {
      return i != listener;
    });
  }
};

/**
 * This is a private API for the startup and platform integration code to use
 * It is not properly documented and subject to change at any point without
 * warning. Should not be used outside of core Mozilla code.
 */
var AddonManagerInternal = {
  startup: function() {
    AddonManagerPrivate.startup();
  },

  registerProvider: function(provider) {
    AddonManagerPrivate.registerProvider(provider);
  },

  shutdown: function() {
    AddonManagerPrivate.shutdown();
  },

  backgroundUpdateCheck: function() {
    AddonManagerPrivate.backgroundUpdateCheck();
  },

  notifyAddonSelected: function(id, type, pendingRestart) {
    AddonManagerPrivate.notifyAddonSelected(id, type, pendingRestart);
  },

  callInstallListeners: function(method) {
    return AddonManagerPrivate.callInstallListeners.apply(AddonManagerPrivate, arguments);
  },

  callAddonListeners: function(method) {
    AddonManagerPrivate.callAddonListeners.apply(AddonManagerPrivate, arguments);
  }
};

/**
 * This is the public API that UI and developers should be calling
 */
var AddonManager = {
  STATE_AVAILABLE: 0,
  STATE_DOWNLOADING: 1,
  STATE_CHECKING: 2,
  STATE_DOWNLOADED: 3,
  STATE_DOWNLOAD_FAILED: 4,
  STATE_INSTALLING: 5,
  STATE_INSTALLED: 6,
  STATE_INSTALL_FAILED: 7,
  STATE_CANCELLED: 8,

  INSTALLERROR_INVALID_VERSION: -1,
  INSTALLERROR_INVALID_GUID: -2,
  INSTALLERROR_INCOMPATIBLE_VERSION: -3,
  INSTALLERROR_INCOMPATIBLE_PLATFORM: -5,
  INSTALLERROR_BLOCKLISTED: -6,
  INSTALLERROR_INSECURE_UPDATE: -7,
  INSTALLERROR_INVALID_MANIFEST: -8,
  INSTALLERROR_RESTRICTED: -9,
  INSTALLERROR_SOFTBLOCKED: -10,

  DOWNLOADERROR_NETWORK: -1,
  DOWNLOADERROR_INCORRECT_HASH: -2,
  DOWNLOADERROR_CORRUPT_FILE: -3,

  UPDATE_WHEN_USER_REQUESTED: 1,
  UPDATE_WHEN_NEW_APP_DETECTED: 2,
  UPDATE_WHEN_NEW_APP_INSTALLED: 3,
  UPDATE_WHEN_PERIODIC_UPDATE: 16,
  UPDATE_WHEN_ADDON_INSTALLED: 17,

  PENDING_ENABLE: 1,
  PENDING_DISABLE: 2,
  PENDING_UNINSTALL: 4,
  PENDING_INSTALL: 8,

  PERM_CANUNINSTALL: 1,
  PERM_CANENABLE: 2,
  PERM_CANDISABLE: 4,
  PERM_CANUPGRADE: 8,

  getInstallForURL: function(url, callback, mimetype, hash, name, iconURL, version, loadgroup) {
    AddonManagerPrivate.getInstallForURL(url, callback, mimetype, hash, name, iconURL, version, loadgroup);
  },

  getInstallForFile: function(file, callback, mimetype) {
    AddonManagerPrivate.getInstallForFile(file, callback, mimetype);
  },

  getAddon: function(id, callback) {
    AddonManagerPrivate.getAddon(id, callback);
  },

  getAddons: function(ids, callback) {
    AddonManagerPrivate.getAddons(ids, callback);
  },

  getAddonsWithPendingOperations: function(types, callback) {
    AddonManagerPrivate.getAddonsWithPendingOperations(types, callback);
  },

  getAddonsByType: function(types, callback) {
    AddonManagerPrivate.getAddonsByType(types, callback);
  },

  getInstalls: function(types, callback) {
    AddonManagerPrivate.getInstalls(types, callback);
  },

  isInstallEnabled: function(type) {
    return AddonManagerPrivate.isInstallEnabled(type);
  },

  isInstallAllowed: function(type, uri) {
    return AddonManagerPrivate.isInstallAllowed(type, uri);
  },

  startInstallation: function(type, source, uri, installs) {
    AddonManagerPrivate.startInstallation(type, source, uri, installs);
  },

  addInstallListener: function(listener) {
    AddonManagerPrivate.addInstallListener(listener);
  },

  removeInstallListener: function(listener) {
    AddonManagerPrivate.removeInstallListener(listener);
  },

  addAddonListener: function(listener) {
    AddonManagerPrivate.addAddonListener(listener);
  },

  removeAddonListener: function(listener) {
    AddonManagerPrivate.removeAddonListener(listener);
  }
};
