/*
*/

const Cc = Components.classes;
const Ci = Components.interfaces;

var EXPORTED_SYMBOLS = [ "UpdateChecker" ];

const TIMEOUT               = 2 * 60 * 1000;
const PREFIX_NS_RDF         = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
const PREFIX_NS_EM          = "http://www.mozilla.org/2004/em-rdf#";
const PREFIX_ITEM           = "urn:mozilla:item:";
const PREFIX_EXTENSION      = "urn:mozilla:extension:";
const PREFIX_THEME          = "urn:mozilla:theme:";
const TOOLKIT_ID            = "toolkit@mozilla.org"
const XMLURI_PARSE_ERROR    = "http://www.mozilla.org/newlayout/xml/parsererror.xml"

// shared code for suppressing bad cert dialogs
Components.utils.import("resource://gre/modules/CertUtils.jsm");

var gRDF = Cc["@mozilla.org/rdf/rdf-service;1"].
           getService(Ci.nsIRDFService);
var gVersionComparator = Cc["@mozilla.org/xpcom/version-comparator;1"].
                         getService(Ci.nsIVersionComparator);
// Defer getting this since it may not be defined at the start of unit tests
var gAppInfo;

function LOG(str) {
  dump("*** addons.updates: " + str + "\n");
}

function parseRDFManifest(id, type, request) {
  function EM_R(prop) {
    return gRDF.GetResource(PREFIX_NS_EM + prop);
  }

  function getValue(literal) {
    if (literal instanceof Ci.nsIRDFLiteral)
      return literal.Value;
    if (literal instanceof Ci.nsIRDFResource)
      return literal.Value;
    if (literal instanceof Ci.nsIRDFInt)
      return literal.Value;
    return null;
  }

  function getProperty(ds, source, property) {
    return getValue(ds.GetTarget(source, EM_R(property), true));
  }

  function getRequiredProperty(ds, source, property) {
    let value = getProperty(ds, source, property);
    if (!value)
      throw "Missing required property " + property;
    return value;
  }

  let rdfParser = Cc["@mozilla.org/rdf/xml-parser;1"].
                  createInstance(Ci.nsIRDFXMLParser);
  let ds = Cc["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"].
           createInstance(Ci.nsIRDFDataSource);
  rdfParser.parseString(ds, request.channel.URI, request.responseText);

  let item = id;
  switch (type) {
  case "extension":
    item = PREFIX_EXTENSION + item;
    break;
  case "theme":
    item = PREFIX_THEME + item;
    break;
  default:
    item = PREFIX_ITEM + item;
    break;
  }

  // TODO check signature
  let extensionRes  = gRDF.GetResource(item);
  let updates = ds.GetTarget(extensionRes, EM_R("updates"), true);

  if (!updates || !(updates instanceof Ci.nsIRDFResource))
    throw "Missing updates property for " + extensionRes.Value;

  let cu = Cc["@mozilla.org/rdf/container-utils;1"].
           getService(Ci.nsIRDFContainerUtils);
  if (!cu.IsContainer(ds, updates))
    throw "Updates property was not an RDF container";

  let results = [];
  let ctr = Cc["@mozilla.org/rdf/container;1"].
            createInstance(Ci.nsIRDFContainer);
  ctr.Init(ds, updates);
  let items = ctr.GetElements();
  while (items.hasMoreElements()) {
    let item = items.getNext().QueryInterface(Ci.nsIRDFResource);
    let version = getRequiredProperty(ds, item, "version");
    LOG("Found an update entry for version " + version);

    let targetApps = ds.GetTargets(item, EM_R("targetApplication"), true);
    while (targetApps.hasMoreElements()) {
      let targetApp = targetApps.getNext().QueryInterface(Ci.nsIRDFResource);

      let result = {
        version: version,
        updateURL: getProperty(ds, targetApp, "updateLink"),
        updateHash: getProperty(ds, targetApp, "updateHash"),
        updateInfoURL: getProperty(ds, targetApp, "updateInfoURL"),
        targetApplications: [{
          id: getRequiredProperty(ds, targetApp, "id"),
          minVersion: getRequiredProperty(ds, targetApp, "minVersion"),
          maxVersion: getRequiredProperty(ds, targetApp, "maxVersion"),
        }]
      };
      // TODO Strip the link and hash if not secure
      results.push(result);
    }
  }
  return results;
}

function UpdateParser(id, type, url, observer) {
  this.id = id;
  this.type = type;
  this.observer = observer;

  this.timer = Cc["@mozilla.org/timer;1"].
               createInstance(Ci.nsITimer);
  this.timer.initWithCallback(this, TIMEOUT, Ci.nsITimer.TYPE_ONE_SHOT);

  LOG("Requesting " + url);
  this.request = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].
                 createInstance(Ci.nsIXMLHttpRequest);
  this.request.open("GET", url, true);
  this.request.channel.notificationCallbacks = new BadCertHandler();
  this.request.overrideMimeType("text/xml");
  var self = this;
  this.request.onload = function(event) { self.onLoad() };
  this.request.onerror = function(event) { self.onError() };
  this.request.send(null);
}

UpdateParser.prototype = {
  id: null,
  type: null,
  observer: null,
  request: null,
  timer: null,

  onLoad: function() {
    this.timer.cancel();
    this.timer = null;
    let request = this.request;
    this.request = null;

    try {
      checkCert(request.channel);
    }
    catch (e) {
      this.notifyError(UpdateChecker.ERROR_SECURITY_ERROR);
      return;
    }

    if (request.status != 200 && request.status != 0) {
      LOG("Request failed: " + request.status);
      this.notifyError(UpdateChecker.ERROR_DOWNLOAD_ERROR);
      return;
    }

    let xml = request.responseXML;
    if (!xml || xml.documentElement.namespaceURI == XMLURI_PARSE_ERROR) {
      LOG("Update manifest was not valid XML");
      this.notifyError(UpdateChecker.ERROR_PARSE_ERROR);
      return;
    }

    if (xml.documentElement.namespaceURI == PREFIX_NS_RDF) {
      let results = null;
      try {
        results = parseRDFManifest(this.id, this.type, request);
      }
      catch (e) {
        LOG("Error parsing RDF updates: " + e);
        this.notifyError(UpdateChecker.ERROR_PARSE_ERROR);
        return;
      }
      if ("onUpdateCheckComplete" in this.observer)
        this.observer.onUpdateCheckComplete(results);
      return;
    }

    LOG("Update manifest had an unrecognised namespace: " + xml.documentElement.namespaceURI);
    this.notifyError(UpdateChecker.ERROR_UNKNOWN_FORMAT);
  },

  onError: function() {
    if (!this.timer)
      return;

    this.timer.cancel();
    this.timer = null;

    LOG("Request failed: " + this.request.status);
    this.request = null;

    this.notifyError(UpdateChecker.ERROR_DOWNLOAD_ERROR);
  },

  notifyError: function(status) {
    if ("onUpdateCheckFailed" in this.observer)
      this.observer.onUpdateCheckFailed(status);
  },

  notify: function(timer) {
    this.timer = null;
    this.request.abort();
    this.request = null;

    LOG("Request timed out");

    this.notifyError(UpdateChecker.ERROR_TIMEOUT);
  }
};

function matchesApplication(update, appVersion, platformVersion) {
  let result = false;
  for (let i = 0; i < update.targetApplications.length; i++) {
    let app = update.targetApplications[i];
    if (app.id == gAppInfo.ID) {
      return (gVersionComparator.compare(appVersion, app.minVersion) >= 0) &&
             (gVersionComparator.compare(appVersion, app.maxVersion) <= 0);
    }
    if (app.id == TOOLKIT_ID) {
      result = (gVersionComparator.compare(platformVersion, app.minVersion) >= 0) &&
               (gVersionComparator.compare(platformVersion, app.maxVersion) <= 0);
    }
  }
  return result;
}

var UpdateChecker = {
  ERROR_TIMEOUT: -1,
  ERROR_DOWNLOAD_ERROR: -2,
  ERROR_PARSE_ERROR: -3,
  ERROR_UNKNOWN_FORMAT: -4,
  ERROR_SECURITY_ERROR: -5,

  getCompatibilityUpdate: function(updates, version, appVersion, platformVersion) {
    if (!gAppInfo) {
      gAppInfo = Cc["@mozilla.org/xre/app-info;1"].
                 getService(Ci.nsIXULAppInfo);
    }
    if (!appVersion)
      appVersion = gAppInfo.version;
    if (!platformVersion)
      platformVersion = gAppInfo.platformVersion;

    for (let i = 0; i < updates.length; i++) {
      if ((gVersionComparator.compare(updates[i].version, version) == 0) &&
          matchesApplication(updates[i], appVersion, platformVersion))
        return updates[i];
    }
    return null;
  },

  getNewestCompatibleUpdate: function(updates, appVersion, platformVersion) {
    if (!gAppInfo) {
      gAppInfo = Cc["@mozilla.org/xre/app-info;1"].
                 getService(Ci.nsIXULAppInfo);
    }
    if (!appVersion)
      appVersion = gAppInfo.version;
    if (!platformVersion)
      platformVersion = gAppInfo.platformVersion;

    let newest = null;
    for (let i = 0; i < updates.length; i++) {
      if (!updates[i].updateURL)
        continue;
      if ((newest == null || (gVersionComparator.compare(newest.version, updates[i].version) < 0)) &&
          matchesApplication(updates[i], appVersion, platformVersion))
        newest = updates[i];
    }
    return newest;
  },

  checkForUpdates: function(id, type, url, observer) {
    new UpdateParser(id, type, url, observer);
  }
};
